# -*- encoding: utf-8

import pytest

from gearman import connection, compat
from gearman.errors import ConnectionError, ServerUnavailable
from gearman.protocol import GEARMAN_COMMAND_TEXT_COMMAND, GEARMAN_COMMAND_ECHO_REQ


def test_no_host_is_ServerUnavailable():
    with pytest.raises(ServerUnavailable):
        connection.GearmanConnection(host=None)


@pytest.mark.parametrize('keyfile,certfile,ca_certs,expected_use_ssl', [
    (None, None, None, False),
    ('key.txt', None, None, False),
    (None, 'cert.txt', None, False),
    (None, None, 'ca_certs.txt', False),
    ('key.txt', 'cert.txt', None, False),
    (None, 'cert.txt', 'ca_certs.txt', False),
    ('key.txt', None, 'ca_certs.txt', False),
    ('key.txt', 'cert.txt', 'ca_certs.txt', True)
])
def test_use_ssl_only_if_all_three_files(keyfile, certfile, ca_certs, expected_use_ssl):
    conn = connection.GearmanConnection(
        host='localhost',
        keyfile=keyfile,
        certfile=certfile,
        ca_certs=ca_certs
    )
    assert conn.use_ssl == expected_use_ssl


def test_no_socket_means_no_fileno():
    conn = connection.GearmanConnection(host='localhost')
    with pytest.raises(ConnectionError, match='no socket set'):
        conn.fileno()


def test_send_commands_to_buffer():
    conn = connection.GearmanConnection(host='localhost')
    assert conn.send_commands_to_buffer() is None
    assert conn._outgoing_buffer == b''
    conn._outgoing_commands.append((GEARMAN_COMMAND_ECHO_REQ, {"data": "test"}))
    conn.send_commands_to_buffer()
    assert conn._outgoing_buffer == b"\x00REQ\x00\x00\x00\x10\x00\x00\x00\x04test"
    if compat.PY3:
        assert isinstance(conn._outgoing_buffer, compat.binary_type)
    else:
        assert isinstance(conn._outgoing_buffer, compat.binary_type)
    conn._reset_connection()
    conn._outgoing_commands.append((GEARMAN_COMMAND_TEXT_COMMAND, {"raw_text": "raw---text"}))
    conn.send_commands_to_buffer()
    assert conn._outgoing_buffer == b"raw---text"
    if compat.PY3:
        assert isinstance(conn._outgoing_buffer, compat.binary_type)
    else:
        assert isinstance(conn._outgoing_buffer, compat.binary_type)



# -*- encoding: utf-8

import errno
import logging

from . import compat
import gearman.io
import gearman.util
from gearman.connection import GearmanConnection
from gearman.errors import ConnectionError, GearmanError, ServerUnavailable
from gearman.job import GearmanJob

gearman_logger = logging.getLogger(__name__)


class DataEncoder(object):
    @classmethod
    def encode(cls, encodable_object):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def decode(cls, decodable_string):  # pragma: no cover
        raise NotImplementedError


class NoopEncoder(DataEncoder):
    """Provide common object dumps for all communications over gearman"""
    @classmethod
    def _enforce_byte_string(cls, given_object):
        if not isinstance(given_object, compat.binary_type):
            if not isinstance(given_object, str):
                # Postel's: Provide Python 2 => Python 3 compatibility.
                raise TypeError("Expecting byte string, got %r" % type(given_object))

    @classmethod
    def encode(cls, encodable_object):
        cls._enforce_byte_string(encodable_object)
        return encodable_object

    @classmethod
    def decode(cls, decodable_string):
        cls._enforce_byte_string(decodable_string)
        return decodable_string


class GearmanConnectionManager(object):
    """Abstract base class for any Gearman-type client that needs to connect/listen to multiple connections

    Mananges and polls a group of gearman connections
    Forwards all communication between a connection and a command handler
    The state of a connection is represented within the command handler

    Automatically encodes all 'data' fields as specified in protocol.py
    """
    command_handler_class = None
    connection_class = GearmanConnection

    job_class = GearmanJob

    data_encoder = NoopEncoder

    def __init__(self, host_list=None):
        assert self.command_handler_class is not None, 'GearmanClientBase did not receive a command handler class'

        self.connection_list = []

        host_list = host_list or []
        for element in host_list:
            # old style host:port pair
            if isinstance(element, (str, tuple)):
                self.add_connection(element)
            elif isinstance(element, dict):
                if not all(k in element for k in ('host', 'port', 'keyfile', 'certfile', 'ca_certs')):
                    raise GearmanError("Incomplete SSL connection definition")
                self.add_ssl_connection(element['host'], element['port'],
                                        element['keyfile'], element['certfile'],
                                        element['ca_certs'])
            else:
                raise GearmanError(
                    "Expected str, tuple or dict; got %r (type %s)" %
                    (element, type(element))
                )

        self.handler_to_connection_map = {}
        self.connection_to_handler_map = {}

        self.handler_initial_state = {}

    def __repr__(self):
        return '<%s connection_list=%r>' % (
            type(self).__name__, self.connection_list)

    def shutdown(self):
        # Shutdown all our connections one by one
        for gearman_connection in self.connection_list:
            gearman_connection.close()

    ###################################
    # Connection management functions #
    ###################################

    def add_ssl_connection(self, host, port, keyfile, certfile, ca_certs):
        """Add a new SSL connection to this connection manager"""
        client_connection = self.connection_class(host=host,
                                                  port=port,
                                                  keyfile=keyfile,
                                                  certfile=certfile,
                                                  ca_certs=ca_certs)
        self.connection_list.append(client_connection)
        return client_connection

    def add_connection(self, hostport_tuple):
        """Add a new connection to this connection manager"""
        gearman_host, gearman_port = gearman.util.disambiguate_server_parameter(hostport_tuple)

        client_connection = self.connection_class(host=gearman_host, port=gearman_port)
        self.connection_list.append(client_connection)

        return client_connection

    def establish_connection(self, current_connection):
        """Attempt to connect... if not previously connected, create a new CommandHandler to manage this connection's state
        !NOTE! This function can throw a ConnectionError which deriving ConnectionManagers should catch
        """
        assert current_connection in self.connection_list, "Unknown connection - %r" % current_connection
        if current_connection.connected:
            return current_connection

        # !NOTE! May throw a ConnectionError
        current_connection.connect()

        # Initiate a new command handler every time we start a new connection
        current_handler = self.command_handler_class(connection_manager=self)

        # Handler to connection map for CommandHandler -> Connection interactions
        # Connection to handler map for Connection -> CommandHandler interactions
        self.handler_to_connection_map[current_handler] = current_connection
        self.connection_to_handler_map[current_connection] = current_handler

        current_handler.initial_state(**self.handler_initial_state)
        return current_connection

    def poll_connections_once(self, poller, connection_map, timeout=None):
        # a timeout of -1 when used with epoll will block until there
        # is activity. Select does not support negative timeouts, so this
        # is translated to a timeout=None when falling back to select
        timeout = timeout or -1

        readable = set()
        writable = set()
        errors = set()
        for fileno, events in poller.poll(timeout=timeout):
            conn = connection_map.get(fileno)
            if not conn:
                continue
            if events & gearman.io.READ:
                readable.add(conn)
            if events & gearman.io.WRITE:
                writable.add(conn)
            if events & gearman.io.ERROR:
                errors.add(conn)

        return readable, writable, errors

    def handle_connection_activity(self, rd_connections, wr_connections, ex_connections):
        """Process all connection activity... executes all handle_* callbacks"""
        dead_connections = set()
        for current_connection in rd_connections:
            try:
                self.handle_read(current_connection)
            except ConnectionError:
                dead_connections.add(current_connection)

        for current_connection in wr_connections:
            try:
                self.handle_write(current_connection)
            except ConnectionError:
                dead_connections.add(current_connection)

        for current_connection in ex_connections:
            self.handle_error(current_connection)

        for current_connection in dead_connections:
            self.handle_error(current_connection)

        failed_connections = ex_connections | dead_connections
        return rd_connections, wr_connections, failed_connections

    def _register_connections_with_poller(self, connections, poller):
        for conn in connections:
            # possible that not all connections have been established yet
            if not conn.gearman_socket:
                continue
            try:
                poller.unregister(conn)
            except IOError as exc:
                if exc.errno != errno.ENOENT:
                    raise
            events = 0
            if conn.readable():
                events |= gearman.io.READ
            if conn.writable():
                events |= gearman.io.WRITE
            poller.register(conn, events)

    def poll_connections_until_stopped(self, submitted_connections, callback_fxn, timeout=None):
        """Continue to poll our connections until we receive a stopping condition"""
        stopwatch = gearman.util.Stopwatch(timeout)
        submitted_connections = set(submitted_connections)
        connection_map = {}

        any_activity = False
        callback_ok = callback_fxn(any_activity)
        connection_ok = any(current_connection.connected for current_connection in submitted_connections)
        poller = gearman.io.get_connection_poller()
        if connection_ok:
            connection_map = {
                conn.fileno(): conn
                for conn in submitted_connections
                if conn.connected
            }

        while connection_ok and callback_ok:
            self._register_connections_with_poller(submitted_connections, poller)

            time_remaining = stopwatch.get_time_remaining()
            if time_remaining == 0.0:
                break

            # Do a single robust select and handle all connection activity
            read_connections, write_connections, dead_connections = self.poll_connections_once(poller, connection_map, timeout=time_remaining)

            # Handle reads and writes and close all of the dead connections
            read_connections, write_connections, dead_connections = self.handle_connection_activity(read_connections, write_connections, dead_connections)

            any_activity = any([read_connections, write_connections, dead_connections])

            # Do not retry dead connections on the next iteration of the loop, as we closed them in handle_error
            submitted_connections -= dead_connections

            callback_ok = callback_fxn(any_activity)
            connection_ok = any(current_connection.connected for current_connection in submitted_connections)

        poller.close()

        # We should raise here if we have no alive connections (don't go into a select polling loop with no connections)
        if not connection_ok:
            raise ServerUnavailable('Found no valid connections in list: %r' % self.connection_list)

        return bool(connection_ok and callback_ok)

    def handle_read(self, current_connection):
        """Handle all our pending socket data"""
        current_handler = self.connection_to_handler_map[current_connection]

        # Transfer data from socket -> buffer
        current_connection.read_data_from_socket()

        # Transfer command from buffer -> command queue
        current_connection.read_commands_from_buffer()

        # Notify the handler that we have commands to fetch
        current_handler.fetch_commands()

    def handle_write(self, current_connection):
        # Transfer command from command queue -> buffer
        current_connection.send_commands_to_buffer()

        # Transfer data from buffer -> socket
        current_connection.send_data_to_socket()

    def handle_error(self, current_connection):
        dead_handler = self.connection_to_handler_map.pop(current_connection, None)
        if dead_handler:
            dead_handler.on_io_error()

        self.handler_to_connection_map.pop(dead_handler, None)
        current_connection.close()

    ##################################
    # Callbacks for Command Handlers #
    ##################################

    def read_command(self, command_handler):
        """CommandHandlers call this function to fetch pending commands

        NOTE: CommandHandlers have NO knowledge as to which connection they're representing
              ConnectionManagers must forward inbound commands to CommandHandlers
        """
        gearman_connection = self.handler_to_connection_map[command_handler]
        cmd_tuple = gearman_connection.read_command()
        if cmd_tuple is None:
            return cmd_tuple

        cmd_type, cmd_args = cmd_tuple
        return cmd_type, cmd_args

    def send_command(self, command_handler, cmd_type, cmd_args):
        """CommandHandlers call this function to send pending commands

        NOTE: CommandHandlers have NO knowledge as to which connection they're representing
              ConnectionManagers must forward outbound commands to Connections
        """
        gearman_connection = self.handler_to_connection_map[command_handler]
        gearman_connection.send_command(cmd_type, cmd_args)

    def on_gearman_error(self, error_code, error_text):
        gearman_logger.error('Received error from server: %s: %s' % (error_code, error_text))
        return False



# -*- encoding: utf-8

import random
import sys

from gearman.connection_manager import GearmanConnectionManager
from gearman.worker_handler import GearmanWorkerCommandHandler
from gearman.errors import ConnectionError

POLL_TIMEOUT_IN_SECONDS = 60.0


class GearmanWorker(GearmanConnectionManager):
    """
    GearmanWorker :: Interface to accept jobs from a Gearman server
    """
    command_handler_class = GearmanWorkerCommandHandler

    def __init__(self, host_list=None):
        super(GearmanWorker, self).__init__(host_list=host_list)

        self.randomized_connections = None

        self.worker_abilities = {}
        self.worker_client_id = None
        self.command_handler_holding_job_lock = None

        self._update_initial_state()

    def _update_initial_state(self):
        self.handler_initial_state['abilities'] = self.worker_abilities.keys()
        self.handler_initial_state['client_id'] = self.worker_client_id

    ########################################################
    ##### Public methods for general GearmanWorker use #####
    ########################################################
    def register_task(self, task, callback_function):
        """Register a function with this worker

        def function_callback(calling_gearman_worker, current_job):
            return current_job.data
        """
        self.worker_abilities[task] = callback_function
        self._update_initial_state()

        for command_handler in self.handler_to_connection_map:
            command_handler.set_abilities(self.handler_initial_state['abilities'])

        return task

    def unregister_task(self, task):
        """Unregister a function with worker"""
        self.worker_abilities.pop(task, None)
        self._update_initial_state()

        for command_handler in self.handler_to_connection_map:
            command_handler.set_abilities(self.handler_initial_state['abilities'])

        return task

    def set_client_id(self, client_id):
        """Notify the server that we should be identified as this client ID"""
        self.worker_client_id = client_id
        self._update_initial_state()

        for command_handler in self.handler_to_connection_map:
            command_handler.set_client_id(self.handler_initial_state['client_id'])

        return client_id

    def work(self, poll_timeout=POLL_TIMEOUT_IN_SECONDS):
        """Loop indefinitely, complete tasks from all connections."""
        continue_working = True
        worker_connections = []

        # We're going to track whether a previous call to our closure indicated
        # we were processing a job. This is just a list of possibly a single
        # element indicating we had a job. It's a list so that through the
        # magic of closures we can reference and write to it each call.
        # This is all so that we can determine when we've finished processing a job
        # correctly.
        had_job = []

        def continue_while_connections_alive(any_activity):
            if had_job and not self.has_job_lock():
                return self.after_poll(any_activity) and self.after_job()

            del had_job[:]
            if self.has_job_lock():
                had_job.append(True)

            return self.after_poll(any_activity)

        # Shuffle our connections after the poll timeout
        while continue_working:
            worker_connections = self.establish_worker_connections()
            continue_working = self.poll_connections_until_stopped(worker_connections, continue_while_connections_alive, timeout=poll_timeout)

        # If we were kicked out of the worker loop, we should shutdown all our connections
        for current_connection in worker_connections:
            current_connection.close()

    def shutdown(self):
        self.command_handler_holding_job_lock = None
        super(GearmanWorker, self).shutdown()

    ###############################################################
    ## Methods to override when dealing with connection polling ##
    ##############################################################
    def establish_worker_connections(self):
        """Return a shuffled list of connections that are alive, and try to reconnect to dead connections if necessary."""
        self.randomized_connections = list(self.connection_list)
        random.shuffle(self.randomized_connections)

        output_connections = []
        for current_connection in self.randomized_connections:
            try:
                valid_connection = self.establish_connection(current_connection)
                output_connections.append(valid_connection)
            except ConnectionError:
                pass

        return output_connections

    def after_poll(self, any_activity):
        """Polling callback to notify any outside listeners whats going on with the GearmanWorker.

        Return True to continue polling, False to exit the work loop"""
        return True

    def after_job(self):
        """Callback to notify any outside listeners that a GearmanWorker has completed the current job.

        This is useful for accomplishing work or stopping the GearmanWorker in between jobs.

        Return True to continue polling, False to exit the work loop
        """
        return True

    def handle_error(self, current_connection):
        """If we discover that a connection has a problem, we better release the job lock"""
        current_handler = self.connection_to_handler_map.get(current_connection)
        if current_handler:
            self.set_job_lock(current_handler, lock=False)

        super(GearmanWorker, self).handle_error(current_connection)

    #############################################################
    ## Public methods so Gearman jobs can send Gearman updates ##
    #############################################################
    def _get_handler_for_job(self, current_job):
        return self.connection_to_handler_map[current_job.connection]

    def wait_until_updates_sent(self, multiple_gearman_jobs, poll_timeout=None):
        connection_set = set([current_job.connection for current_job in multiple_gearman_jobs])

        def continue_while_updates_pending(any_activity):
            return any(current_connection.writable() for current_connection in connection_set)

        self.poll_connections_until_stopped(connection_set, continue_while_updates_pending, timeout=poll_timeout)

    def send_job_status(self, current_job, numerator, denominator, poll_timeout=None):
        """Send a Gearman JOB_STATUS update for an inflight job"""
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_status(current_job, numerator=numerator, denominator=denominator)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    def send_job_complete(self, current_job, data, poll_timeout=None):
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_complete(current_job, data=data)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    def send_job_failure(self, current_job, poll_timeout=None):
        """Removes a job from the queue if its backgrounded"""
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_failure(current_job)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    def send_job_exception(self, current_job, data, poll_timeout=None):
        """Removes a job from the queue if its backgrounded"""
        # Using GEARMAND_COMMAND_WORK_EXCEPTION is not recommended at time of this writing [2010-02-24]
        # http://groups.google.com/group/gearman/browse_thread/thread/5c91acc31bd10688/529e586405ed37fe
        #
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_exception(current_job, data=data)
        current_handler.send_job_failure(current_job)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    def send_job_data(self, current_job, data, poll_timeout=None):
        """Send a Gearman JOB_DATA update for an inflight job"""
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_data(current_job, data=data)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    def send_job_warning(self, current_job, data, poll_timeout=None):
        """Send a Gearman JOB_WARNING update for an inflight job"""
        current_handler = self._get_handler_for_job(current_job)
        current_handler.send_job_warning(current_job, data=data)

        self.wait_until_updates_sent([current_job], poll_timeout=poll_timeout)

    #####################################################
    ##### Callback methods for GearmanWorkerHandler #####
    #####################################################
    def create_job(self, command_handler, job_handle, task, unique, data):
        """Create a new job using our self.job_class"""
        current_connection = self.handler_to_connection_map[command_handler]
        return self.job_class(current_connection, job_handle, task, unique, data)

    def on_job_execute(self, current_job):
        try:
            function_callback = self.worker_abilities[current_job.task]
            job_result = function_callback(self, current_job)
        except Exception:
            return self.on_job_exception(current_job, sys.exc_info())

        return self.on_job_complete(current_job, job_result)

    def on_job_exception(self, current_job, exc_info):
        self.send_job_failure(current_job)
        return False

    def on_job_complete(self, current_job, job_result):
        self.send_job_complete(current_job, job_result)
        return True

    def set_job_lock(self, command_handler, lock):
        """Set a worker level job lock so we don't try to hold onto 2 jobs at anytime"""
        if command_handler not in self.handler_to_connection_map:
            return False

        failed_lock = bool(lock and self.command_handler_holding_job_lock is not None)
        failed_unlock = bool(not lock and self.command_handler_holding_job_lock != command_handler)

        # If we've already been locked, we should say the lock failed
        # If we're attempting to unlock something when we don't have a lock, we're in a bad state
        if failed_lock or failed_unlock:
            return False

        if lock:
            self.command_handler_holding_job_lock = command_handler
        else:
            self.command_handler_holding_job_lock = None

        return True

    def has_job_lock(self):
        return bool(self.command_handler_holding_job_lock is not None)

    def check_job_lock(self, command_handler):
        """Check to see if we hold the job lock"""
        return bool(self.command_handler_holding_job_lock == command_handler)



# -*- encoding: utf-8

import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


if PY3:
    binary_type = bytes
    unicode_type = str

    def array_to_bytes(arr):
        return arr.tobytes()

    def itervalues(d):
        return d.values()

    def to_hex(binary_str):
        return binary_str.hex()

else:
    binary_type = str
    unicode_type = unicode

    def array_to_bytes(arr):
        return arr.tostring()

    def itervalues(d):
        return d.itervalues()

    def to_hex(binary_str):
        return binary_str.encode('hex')



# -*- encoding: utf-8

import collections

import pytest

from gearman.client import GearmanClient
from gearman.client_handler import GearmanClientCommandHandler

from gearman.constants import PRIORITY_NONE, PRIORITY_HIGH, PRIORITY_LOW, JOB_UNKNOWN, JOB_PENDING, JOB_CREATED, JOB_FAILED, JOB_COMPLETE
from gearman.errors import ExceededConnectionAttempts, ServerUnavailable, InvalidClientState
from gearman.protocol import submit_cmd_for_background_priority, GEARMAN_COMMAND_STATUS_RES, GEARMAN_COMMAND_GET_STATUS, GEARMAN_COMMAND_JOB_CREATED, \
    GEARMAN_COMMAND_WORK_STATUS, GEARMAN_COMMAND_WORK_FAIL, GEARMAN_COMMAND_WORK_COMPLETE, GEARMAN_COMMAND_WORK_DATA, GEARMAN_COMMAND_WORK_WARNING

from tests._core_testing import (
    _GearmanAbstractTest,
    MockGearmanConnectionManager,
    MockGearmanConnection,
    random_bytes
)


class MockGearmanClient(GearmanClient, MockGearmanConnectionManager):
    pass


class ClientTest(_GearmanAbstractTest):
    """Test the public client interface"""
    connection_manager_class = MockGearmanClient
    command_handler_class = GearmanClientCommandHandler

    def setUp(self):
        super(ClientTest, self).setUp()
        self.original_handle_connection_activity = self.connection_manager.handle_connection_activity

    def tearDown(self):
        super(ClientTest, self).tearDown()
        self.connection_manager.handle_connection_activity = self.original_handle_connection_activity

    def generate_job_request(self, submitted=True, accepted=True):
        current_request = super(ClientTest, self).generate_job_request()
        if submitted or accepted:
            self.connection_manager.establish_request_connection(current_request)
            self.command_handler.send_job_request(current_request)

        if submitted and accepted:
            self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=current_request.job.handle)
            assert current_request.job.handle in self.command_handler.handle_to_request_map

        return current_request

    def test_establish_request_connection_complex(self):
        # Spin up a bunch of imaginary gearman connections
        failed_connection = MockGearmanConnection()
        failed_connection._fail_on_bind = True

        failed_then_retried_connection = MockGearmanConnection()
        failed_then_retried_connection._fail_on_bind = True

        good_connection = MockGearmanConnection()
        good_connection.connect()

        # Register all our connections
        self.connection_manager.connection_list = [failed_connection, failed_then_retried_connection, good_connection]

        # When we first create our request, our client shouldn't know anything about it
        current_request = self.generate_job_request(submitted=False, accepted=False)
        assert current_request not in self.connection_manager.request_to_rotating_connection_queue

        # Make sure that when we start up, we get our good connection
        chosen_connection = self.connection_manager.establish_request_connection(current_request)
        assert chosen_connection == good_connection

        assert not failed_connection.connected
        assert not failed_then_retried_connection.connected
        assert good_connection.connected

        # No state changed so we should still go to the correct connection
        chosen_connection = self.connection_manager.establish_request_connection(current_request)
        assert chosen_connection == good_connection

        # Pretend like our good connection died so we'll need to choose somethign else
        good_connection._reset_connection()
        good_connection._fail_on_bind = True

        failed_then_retried_connection._fail_on_bind = False
        failed_then_retried_connection.connect()

        # Make sure we rotate good_connection and failed_connection out
        chosen_connection = self.connection_manager.establish_request_connection(current_request)
        assert chosen_connection == failed_then_retried_connection
        assert not failed_connection.connected
        assert failed_then_retried_connection.connected
        assert not good_connection.connected

    def test_establish_request_connection_dead(self):
        self.connection_manager.connection_list = []
        self.connection_manager.command_handlers = {}

        current_request = self.generate_job_request(submitted=False, accepted=False)

        # No connections == death
        with pytest.raises(ServerUnavailable):
            self.connection_manager.establish_request_connection(current_request)

        # Spin up a bunch of imaginary gearman connections
        failed_connection = MockGearmanConnection()
        failed_connection._fail_on_bind = True
        self.connection_manager.connection_list.append(failed_connection)

        # All failed connections == death
        with pytest.raises(ServerUnavailable):
            self.connection_manager.establish_request_connection(current_request)

    def test_auto_retry_behavior(self):
        current_request = self.generate_job_request(submitted=False, accepted=False)

        def fail_then_create_jobs(rx_conns, wr_conns, ex_conns):
            if self.connection_manager.current_failures < self.connection_manager.expected_failures:
                self.connection_manager.current_failures += 1

                # We're going to down this connection and reset state
                assert self.connection.connected
                self.connection_manager.handle_error(self.connection)
                assert not self.connection.connected

                # We're then going to IMMEDIATELY pull this connection back up
                # So we don't bail out of the "self.connection_manager.poll_connections_until_stopped" loop
                self.connection_manager.establish_connection(self.connection)
            else:
                assert current_request.state == JOB_PENDING
                self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=current_request.job.handle)

            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = fail_then_create_jobs
        self.connection_manager.expected_failures = 5

        # Now that we've setup our rety behavior, we need to reset the entire state of our experiment
        # First pass should succeed as we JUST touch our max attempts
        self.connection_manager.current_failures = current_request.connection_attempts = 0
        current_request.max_connection_attempts = self.connection_manager.expected_failures + 1
        current_request.state = JOB_UNKNOWN

        self.connection_manager.wait_until_jobs_accepted([current_request])
        assert current_request.state == JOB_CREATED
        assert current_request.connection_attempts == current_request.max_connection_attempts

        # Second pass should fail as we JUST exceed our max attempts
        self.connection_manager.current_failures = current_request.connection_attempts = 0
        current_request.max_connection_attempts = self.connection_manager.expected_failures
        current_request.state = JOB_UNKNOWN

        with pytest.raises(ExceededConnectionAttempts):
            self.connection_manager.wait_until_jobs_accepted([current_request])
        assert current_request.state == JOB_UNKNOWN
        assert current_request.connection_attempts == current_request.max_connection_attempts

    def test_multiple_fg_job_submission(self):
        submitted_job_count = 5
        expected_job_list = [self.generate_job() for _ in range(submitted_job_count)]

        def mark_jobs_created(rx_conns, wr_conns, ex_conns):
            for current_job in expected_job_list:
                self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=current_job.handle)

            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = mark_jobs_created

        job_dictionaries = [current_job.to_dict() for current_job in expected_job_list]

        # Test multiple job submission
        job_requests = self.connection_manager.submit_multiple_jobs(job_dictionaries, wait_until_complete=False)
        for current_request, expected_job in zip(job_requests, expected_job_list):
            current_job = current_request.job
            self.assert_jobs_equal(current_job, expected_job)

            assert current_request.priority == PRIORITY_NONE
            assert not current_request.background
            assert current_request.state == JOB_CREATED

            assert not current_request.complete

    def test_single_bg_job_submission(self):
        expected_job = self.generate_job()

        def mark_job_created(rx_conns, wr_conns, ex_conns):
            self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=expected_job.handle)
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = mark_job_created
        job_request = self.connection_manager.submit_job(expected_job.task, expected_job.data, unique=expected_job.unique, background=True, priority=PRIORITY_LOW, wait_until_complete=False)

        current_job = job_request.job
        self.assert_jobs_equal(current_job, expected_job)

        assert job_request.priority == PRIORITY_LOW
        assert job_request.background
        assert job_request.state == JOB_CREATED

        assert job_request.complete

    def test_single_fg_job_submission_timeout(self):
        expected_job = self.generate_job()

        def job_failed_submission(rx_conns, wr_conns, ex_conns):
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = job_failed_submission
        job_request = self.connection_manager.submit_job(expected_job.task, expected_job.data, unique=expected_job.unique, priority=PRIORITY_HIGH, poll_timeout=0.01)

        assert job_request.priority == PRIORITY_HIGH
        assert not job_request.background
        assert job_request.state == JOB_PENDING

        assert not job_request.complete
        assert job_request.timed_out

    def test_single_fg_job_submission_timeout_unspecified_unique(self):
        expected_job = self.generate_job()

        def job_failed_submission(rx_conns, wr_conns, ex_conns):
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = job_failed_submission
        job_request = self.connection_manager.submit_job(
            expected_job.task,
            expected_job.data,
            unique=None,
            priority=PRIORITY_HIGH,
            poll_timeout=0.01
        )

        assert job_request.priority == PRIORITY_HIGH
        assert not job_request.background
        assert job_request.state == JOB_PENDING

        assert not job_request.complete
        assert job_request.timed_out

    def test_wait_for_multiple_jobs_to_complete_or_timeout(self):
        completed_request = self.generate_job_request()
        failed_request = self.generate_job_request()
        timeout_request = self.generate_job_request()

        self.update_requests = True

        def multiple_job_updates(rx_conns, wr_conns, ex_conns):
            # Only give a single status update and have the 3rd job handle timeout
            if self.update_requests:
                self.command_handler.recv_command(GEARMAN_COMMAND_WORK_COMPLETE, job_handle=completed_request.job.handle, data=b'12345')
                self.command_handler.recv_command(GEARMAN_COMMAND_WORK_FAIL, job_handle=failed_request.job.handle)
                self.update_requests = False

            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = multiple_job_updates

        finished_requests = self.connection_manager.wait_until_jobs_completed([completed_request, failed_request, timeout_request], poll_timeout=0.01)
        del self.update_requests

        finished_completed_request, finished_failed_request, finished_timeout_request = finished_requests

        self.assert_jobs_equal(finished_completed_request.job, completed_request.job)
        assert finished_completed_request.state == JOB_COMPLETE
        assert finished_completed_request.result == b'12345'
        assert not finished_completed_request.timed_out
        # self.assert_(finished_completed_request.job.handle not in self.command_handler.handle_to_request_map)

        self.assert_jobs_equal(finished_failed_request.job, failed_request.job)
        assert finished_failed_request.state == JOB_FAILED
        assert finished_failed_request.result is None
        assert not finished_failed_request.timed_out
        # self.assert_(finished_failed_request.job.handle not in self.command_handler.handle_to_request_map)

        assert finished_timeout_request.state == JOB_CREATED
        assert finished_timeout_request.result is None
        assert finished_timeout_request.timed_out
        assert finished_timeout_request.job.handle in self.command_handler.handle_to_request_map

    def test_get_job_status(self):
        single_request = self.generate_job_request()

        def retrieve_status(rx_conns, wr_conns, ex_conns):
            self.command_handler.recv_command(GEARMAN_COMMAND_STATUS_RES, job_handle=single_request.job.handle, known='1', running='0', numerator='0', denominator='1')
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = retrieve_status

        job_request = self.connection_manager.get_job_status(single_request)
        request_status = job_request.status
        assert request_status
        assert request_status['known']
        assert not request_status['running']
        assert request_status['numerator'] == 0
        assert request_status['denominator'] == 1
        assert not job_request.timed_out

    def test_get_job_status_unknown(self):
        single_request = self.generate_job_request()
        current_handle = single_request.job.handle
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_FAIL, job_handle=current_handle)

        def retrieve_status(rx_conns, wr_conns, ex_conns):
            self.command_handler.recv_command(GEARMAN_COMMAND_STATUS_RES, job_handle=current_handle, known='0', running='0', numerator='0', denominator='1')
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = retrieve_status

        job_request = self.connection_manager.get_job_status(single_request)
        request_status = job_request.status
        assert request_status
        assert not request_status['known']
        assert not request_status['running']
        assert request_status['numerator'] == 0
        assert request_status['denominator'] == 1
        assert not job_request.timed_out
        # self.assert_(current_handle not in self.command_handler.handle_to_request_map)

    def test_get_job_status_timeout(self):
        single_request = self.generate_job_request()

        def retrieve_status_timeout(rx_conns, wr_conns, ex_conns):
            return rx_conns, wr_conns, ex_conns

        self.connection_manager.handle_connection_activity = retrieve_status_timeout

        job_request = self.connection_manager.get_job_status(single_request, poll_timeout=0.01)
        assert job_request.timed_out


class ClientCommandHandlerInterfaceTest(_GearmanAbstractTest):
    """Test the public interface a GearmanClient may need to call in order to update state on a GearmanClientCommandHandler"""
    connection_manager_class = MockGearmanClient
    command_handler_class = GearmanClientCommandHandler

    def test_send_job_request(self):
        current_request = self.generate_job_request()
        gearman_job = current_request.job

        for priority in (PRIORITY_NONE, PRIORITY_HIGH, PRIORITY_LOW):
            for background in (False, True):
                current_request.reset()
                current_request.priority = priority
                current_request.background = background

                self.command_handler.send_job_request(current_request)

                queued_request = self.command_handler.requests_awaiting_handles.popleft()
                assert queued_request == current_request

                expected_cmd_type = submit_cmd_for_background_priority(background, priority)
                self.assert_sent_command(expected_cmd_type, task=gearman_job.task, data=gearman_job.data, unique=gearman_job.unique)

    def test_get_status_of_job(self):
        current_request = self.generate_job_request()

        self.command_handler.send_get_status_of_job(current_request)

        self.assert_sent_command(GEARMAN_COMMAND_GET_STATUS, job_handle=current_request.job.handle)


class ClientCommandHandlerStateMachineTest(_GearmanAbstractTest):
    """Test single state transitions within a GearmanWorkerCommandHandler"""
    connection_manager_class = MockGearmanClient
    command_handler_class = GearmanClientCommandHandler

    def generate_job_request(self, submitted=True, accepted=True):
        current_request = super(ClientCommandHandlerStateMachineTest, self).generate_job_request()
        if submitted or accepted:
            self.command_handler.requests_awaiting_handles.append(current_request)
            current_request.state = JOB_PENDING

        if submitted and accepted:
            self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=current_request.job.handle)

        return current_request

    def test_received_job_created(self):
        current_request = self.generate_job_request(accepted=False)

        new_handle = random_bytes()
        self.command_handler.recv_command(GEARMAN_COMMAND_JOB_CREATED, job_handle=new_handle)

        assert current_request.job.handle == new_handle
        assert current_request.state == JOB_CREATED
        assert self.command_handler.handle_to_request_map[new_handle] == current_request

    def test_received_job_created_out_of_order(self):
        assert self.command_handler.requests_awaiting_handles == collections.deque()

        # Make sure we bail cuz we have an empty queue
        with pytest.raises(InvalidClientState):
            self.command_handler.recv_command(
                GEARMAN_COMMAND_JOB_CREATED,
                job_handle=None
            )

    def test_required_state_pending(self):
        current_request = self.generate_job_request(submitted=False, accepted=False)

        new_handle = random_bytes()

        invalid_states = [JOB_UNKNOWN, JOB_CREATED, JOB_COMPLETE, JOB_FAILED]
        for bad_state in invalid_states:
            current_request.state = bad_state

            # We only want to check the state of request... not die if we don't have any pending requests
            self.command_handler.requests_awaiting_handles.append(current_request)

            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_JOB_CREATED,
                    job_handle=new_handle
                )

    def test_required_state_queued(self):
        current_request = self.generate_job_request()

        job_handle = current_request.job.handle
        new_data = random_bytes()

        invalid_states = [JOB_UNKNOWN, JOB_PENDING, JOB_COMPLETE, JOB_FAILED]
        for bad_state in invalid_states:
            current_request.state = bad_state

            # All these commands expect to be in JOB_CREATED
            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_WORK_DATA,
                    job_handle=job_handle,
                    data=new_data
                )

            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_WORK_WARNING,
                    job_handle=job_handle,
                    data=new_data
                )

            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_WORK_STATUS,
                    job_handle=job_handle,
                    numerator=0,
                    denominator=1
                )

            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_WORK_COMPLETE,
                    job_handle=job_handle,
                    data=new_data
                )

            with pytest.raises(InvalidClientState):
                self.command_handler.recv_command(
                    GEARMAN_COMMAND_WORK_FAIL,
                    job_handle=job_handle
                )

    def test_in_flight_work_updates(self):
        current_request = self.generate_job_request()

        job_handle = current_request.job.handle
        new_data = random_bytes()

        # Test WORK_DATA
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_DATA, job_handle=job_handle, data=new_data)
        assert current_request.data_updates.popleft() == new_data
        assert current_request.state == JOB_CREATED

        # Test WORK_WARNING
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_WARNING, job_handle=job_handle, data=new_data)
        assert current_request.warning_updates.popleft() == new_data
        assert current_request.state == JOB_CREATED

        # Test WORK_STATUS
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_STATUS, job_handle=job_handle, numerator=0, denominator=1)

        assert current_request.state == JOB_CREATED

    def test_work_complete(self):
        current_request = self.generate_job_request()

        job_handle = current_request.job.handle
        new_data = random_bytes()
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_COMPLETE, job_handle=job_handle, data=new_data)

        assert current_request.result == new_data
        assert current_request.state == JOB_COMPLETE

    def test_work_fail(self):
        current_request = self.generate_job_request()

        job_handle = current_request.job.handle
        self.command_handler.recv_command(GEARMAN_COMMAND_WORK_FAIL, job_handle=job_handle)

        assert current_request.state == JOB_FAILED

    def test_status_request(self):
        current_request = self.generate_job_request()

        job_handle = current_request.job.handle

        assert current_request.status == {}

        self.command_handler.recv_command(GEARMAN_COMMAND_STATUS_RES, job_handle=job_handle, known='1', running='1', numerator='0', denominator='1')

        assert current_request.status['handle'] == job_handle
        assert current_request.status['known']
        assert current_request.status['running']
        assert current_request.status['numerator'] == 0
        assert current_request.status['denominator'] == 1



# -*- encoding: utf-8


class GearmanError(Exception):
    pass


class ConnectionError(GearmanError):
    pass


class ServerUnavailable(GearmanError):
    pass


class ProtocolError(GearmanError):
    pass


class UnknownCommandError(GearmanError):
    pass


class ExceededConnectionAttempts(GearmanError):
    pass


class InvalidClientState(GearmanError):
    pass


class InvalidWorkerState(GearmanError):
    pass


class InvalidAdminClientState(GearmanError):
    pass



import select

import gearman.errors
import gearman.util

# epoll event types
_EPOLLIN = 0x01
_EPOLLOUT = 0x04
_EPOLLERR = 0x08
_EPOLLHUP = 0x10

READ = _EPOLLIN
WRITE = _EPOLLOUT
ERROR = _EPOLLERR | _EPOLLHUP


def get_connection_poller():
    """
    Returns a select.epoll-like object. Depending on the platform, this will
    either be:
        - On modern Linux system, with python >= 2.6: select.epoll
        - On all other systems: gearman.io._Select: an object that mimics
          select.epoll, but uses select.select
    """
    if hasattr(select, "epoll"):
        return select.epoll()
    else:
        return _Select()


def _find_bad_connections(connections):
    """
    Find any bad connections in a list of connections.

    For use with select.select.

    When select throws an exception, it's likely that one of the sockets
    passed in has died. In order to find the bad connections, they must be
    checked individually. This will do so and return a list of any bad
    connections found.
    """
    bad = []
    for conn in connections:
        try:
            _, _, _ = gearman.util.select([conn], [], [], timeout=0)
        except (select.error, gearman.errors.ConnectionError):
            bad.append(conn)
    return bad


class _Select(object):
    """
    A `select.epoll`-like object that uses select.select.

    Used as a fallback when epoll is not available. Inspired by tornado's
    fallback mechanism
    """

    def __init__(self):
        self.read = set()
        self.write = set()
        self.error = set()

    def close(self):
        """
        Close the _Select object. For parity with select.epoll. Does nothing
        here.
        """
        pass

    def register(self, fd, evmask):
        """
        Register a file descriptor for polling.

        fd: a file descriptor (socket) to be registers
        evmask: a bit set describing the desired events to report

        Events are similar to those accepted by select.epoll:
            - gearman.io.READ: report when fd is readable (i.e.: a socket.recv
              operation likely won't block, and will yield some data)
            - gearman.io.WRITE: report when fd is writable (i.e.: a socket.send
              operation likely won't block, and will be able to write some
              data)
            - gearman.io.ERROR: report when fd is in an error state
        """
        if fd in self.read or fd in self.write or fd in self.error:
            raise ValueError("Connection already registered: %d" % fd.fileno())
        if evmask & READ:
            self.read.add(fd)
        if evmask & WRITE:
            self.write.add(fd)
        if evmask & ERROR:
            self.error.add(fd)

    def modify(self, fd, evmask):
        """
        Update the IO events that should be reported for a given file
        descriptor. See _Select.register for details on these events
        """
        self.unregister(fd)
        self.register(fd, evmask)

    def unregister(self, fd):
        """
        Stop tracking events for a given file descriptor
        """
        self.read.discard(fd)
        self.write.discard(fd)
        self.error.discard(fd)

    def poll(self, timeout):
        """
        Wait for events for any of the of register file descriptors. The
        maximum time to wait is specified by the timeout value.

        A timeout < 0 will block indefinitely. A timeout of 0 will not block at
        all. And, a timeout > 0 will block for at most that many seconds. The
        timeout parameter may be a floating point number.
        """
        readable = set()
        writable = set()
        errors = set()

        if timeout is not None and timeout < 0.0:
            # for parity with epoll, negative timeout = block until there
            # is activity
            timeout = None

        connections = (self.read | self.write | self.error)

        success = False
        while not success and connections:
            connections -= errors
            try:
                r, w, e = gearman.util.select(
                    self.read,
                    self.write,
                    self.error,
                    timeout
                )
                readable = set(r)
                writable = set(w)
                errors |= set(e)  # this set could already be populated
                success = True
            except (select.error, gearman.errors.ConnectionError):
                bad_conns = _find_bad_connections(connections)
                map(self.read.discard, bad_conns)
                map(self.write.discard, bad_conns)
                map(self.error.discard, bad_conns)
                errors |= set(bad_conns)

        events = {}
        for conn in readable:
            events[conn.fileno()] = events.get(conn.fileno(), 0) | READ
        for conn in writable:
            events[conn.fileno()] = events.get(conn.fileno(), 0) | WRITE
        for conn in errors:
            events[conn.fileno()] = events.get(conn.fileno(), 0) | ERROR

        return events.items()



# -*- encoding: utf-8

import collections

import pytest

from gearman.worker import GearmanWorker
from gearman.worker_handler import GearmanWorkerCommandHandler

from gearman.errors import ServerUnavailable
from gearman.protocol import get_command_name, GEARMAN_COMMAND_RESET_ABILITIES, GEARMAN_COMMAND_CAN_DO, GEARMAN_COMMAND_SET_CLIENT_ID, \
    GEARMAN_COMMAND_NOOP, GEARMAN_COMMAND_PRE_SLEEP, GEARMAN_COMMAND_NO_JOB, GEARMAN_COMMAND_GRAB_JOB_UNIQ, GEARMAN_COMMAND_JOB_ASSIGN_UNIQ, \
    GEARMAN_COMMAND_WORK_STATUS, GEARMAN_COMMAND_WORK_FAIL, GEARMAN_COMMAND_WORK_COMPLETE, GEARMAN_COMMAND_WORK_DATA, GEARMAN_COMMAND_WORK_EXCEPTION, GEARMAN_COMMAND_WORK_WARNING

from tests._core_testing import _GearmanAbstractTest, MockGearmanConnectionManager, MockGearmanConnection


class MockGearmanWorker(MockGearmanConnectionManager, GearmanWorker):
    def __init__(self, *largs, **kwargs):
        super(MockGearmanWorker, self).__init__(*largs, **kwargs)
        self.worker_job_queues = collections.defaultdict(collections.deque)

    def on_job_execute(self, current_job):
        current_handler = self.connection_to_handler_map[current_job.connection]
        self.worker_job_queues[current_handler].append(current_job)


class _GearmanAbstractWorkerTest(_GearmanAbstractTest):
    connection_manager_class = MockGearmanWorker
    command_handler_class = GearmanWorkerCommandHandler

    def setup_command_handler(self):
        super(_GearmanAbstractWorkerTest, self).setup_command_handler()
        self.assert_sent_abilities([])
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)

    def assert_sent_abilities(self, expected_abilities):
        observed_abilities = set()

        self.assert_sent_command(GEARMAN_COMMAND_RESET_ABILITIES)
        for ability in expected_abilities:
            cmd_type, cmd_args = self.connection._outgoing_commands.popleft()

            assert get_command_name(cmd_type) == get_command_name(GEARMAN_COMMAND_CAN_DO)
            observed_abilities.add(cmd_args['task'])

        assert observed_abilities == set(expected_abilities)

    def assert_sent_client_id(self, expected_client_id):
        self.assert_sent_command(GEARMAN_COMMAND_SET_CLIENT_ID, client_id=expected_client_id)


class WorkerTest(_GearmanAbstractWorkerTest):
    """Test the public worker interface"""
    def test_registering_functions(self):
        # Tests that the abilities were set on the GearmanWorker AND the GearmanWorkerCommandHandler
        # Does NOT test that commands were actually sent out as that is tested in GearmanWorkerCommandHandlerInterfaceTest.test_set_abilities
        def fake_callback_one(worker_command_handler, current_job):  # pragma: no cover
            pass

        def fake_callback_two(worker_command_handler, current_job):  # pragma: no cover
            pass

        # Register a single callback
        self.connection_manager.register_task('fake_callback_one', fake_callback_one)
        assert 'fake_callback_one' in self.connection_manager.worker_abilities
        assert 'fake_callback_two' not in self.connection_manager.worker_abilities
        assert self.connection_manager.worker_abilities['fake_callback_one'] == fake_callback_one
        assert set(self.command_handler._handler_abilities) == set(['fake_callback_one'])

        # Register another callback and make sure the command_handler sees the same functions
        self.connection_manager.register_task('fake_callback_two', fake_callback_two)
        assert 'fake_callback_one' in self.connection_manager.worker_abilities
        assert 'fake_callback_two' in self.connection_manager.worker_abilities
        assert self.connection_manager.worker_abilities['fake_callback_one'] == fake_callback_one
        assert self.connection_manager.worker_abilities['fake_callback_two'] == fake_callback_two
        assert set(self.command_handler._handler_abilities) == set(['fake_callback_one', 'fake_callback_two'])

        # Unregister a callback and make sure the command_handler sees the same functions
        self.connection_manager.unregister_task('fake_callback_one')
        assert 'fake_callback_one' not in self.connection_manager.worker_abilities
        assert 'fake_callback_two' in self.connection_manager.worker_abilities
        assert self.connection_manager.worker_abilities['fake_callback_two'] == fake_callback_two
        assert set(self.command_handler._handler_abilities) == set(['fake_callback_two'])

    def test_setting_client_id(self):
        new_client_id = 'HELLO'

        # Make sure nothing is set
        assert self.connection_manager.worker_client_id is None
        assert self.command_handler._client_id is None

        self.connection_manager.set_client_id(new_client_id)

        # Make sure both the client and the connection handler reflect the new state
        assert self.connection_manager.worker_client_id == new_client_id
        assert self.command_handler._client_id == new_client_id

    def test_establish_worker_connections(self):
        self.connection_manager.connection_list = []
        self.connection_manager.command_handlers = {}

        # Spin up a bunch of imaginary gearman connections
        good_connection = MockGearmanConnection()
        good_connection.connect()
        good_connection._fail_on_bind = False

        failed_then_retried_connection = MockGearmanConnection()
        failed_then_retried_connection._fail_on_bind = False

        failed_connection = MockGearmanConnection()
        failed_connection._fail_on_bind = True

        # Register all our connections
        self.connection_manager.connection_list = [good_connection, failed_then_retried_connection, failed_connection]

        # The only alive connections should be the ones that ultimately be connection.connected
        alive_connections = self.connection_manager.establish_worker_connections()
        assert good_connection in alive_connections
        assert failed_then_retried_connection in alive_connections
        assert failed_connection not in alive_connections

    def test_establish_worker_connections_dead(self):
        self.connection_manager.connection_list = []
        self.connection_manager.command_handlers = {}

        # We have no connections so there will never be any work to do
        with pytest.raises(ServerUnavailable):
            self.connection_manager.work()

        # We were started with a dead connection, make sure we bail again
        dead_connection = MockGearmanConnection()
        dead_connection._fail_on_bind = True
        dead_connection.connected = False
        self.connection_manager.connection_list = [dead_connection]

        with pytest.raises(ServerUnavailable):
            self.connection_manager.work()


class WorkerCommandHandlerInterfaceTest(_GearmanAbstractWorkerTest):
    """Test the public interface a GearmanWorker may need to call in order to update state on a GearmanWorkerCommandHandler"""

    def test_on_connect(self):
        expected_abilities = ['function_one', 'function_two', 'function_three']
        expected_client_id = 'my_client_id'

        self.connection.connected = False

        self.connection_manager.set_client_id(expected_client_id)
        self.connection_manager.unregister_task(b'__test_ability__')
        for task in expected_abilities:
            self.connection_manager.register_task(task, None)

        # We were disconnected, connect and wipe pending commands
        self.connection_manager.establish_connection(self.connection)

        # When we attempt a new connection, make sure we get a new command handler
        assert self.command_handler != self.connection_manager.connection_to_handler_map[self.connection]

        self.assert_sent_client_id(expected_client_id)
        self.assert_sent_abilities(expected_abilities)
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)
        self.assert_no_pending_commands()

    def test_set_abilities(self):
        expected_abilities = ['function_one', 'function_two', 'function_three']

        # We were disconnected, connect and wipe pending commands
        self.command_handler.set_abilities(expected_abilities)
        self.assert_sent_abilities(expected_abilities)
        self.assert_no_pending_commands()

    def test_set_client_id(self):
        expected_client_id = 'my_client_id'

        handler_initial_state = {}
        handler_initial_state['abilities'] = []
        handler_initial_state['client_id'] = None

        # We were disconnected, connect and wipe pending commands
        self.command_handler.set_client_id(expected_client_id)
        self.assert_sent_client_id(expected_client_id)
        self.assert_no_pending_commands()

    def test_send_functions(self):
        current_job = self.generate_job()

        # Test GEARMAN_COMMAND_WORK_STATUS
        self.command_handler.send_job_status(current_job, 0, 1)
        self.assert_sent_command(GEARMAN_COMMAND_WORK_STATUS, job_handle=current_job.handle, numerator='0', denominator='1')

        # Test GEARMAN_COMMAND_WORK_COMPLETE
        self.command_handler.send_job_complete(current_job, b'completion data')
        self.assert_sent_command(GEARMAN_COMMAND_WORK_COMPLETE, job_handle=current_job.handle, data=b'completion data')

        # Test GEARMAN_COMMAND_WORK_FAIL
        self.command_handler.send_job_failure(current_job)
        self.assert_sent_command(GEARMAN_COMMAND_WORK_FAIL, job_handle=current_job.handle)

        # Test GEARMAN_COMMAND_WORK_EXCEPTION
        self.command_handler.send_job_exception(current_job, b'exception data')
        self.assert_sent_command(GEARMAN_COMMAND_WORK_EXCEPTION, job_handle=current_job.handle, data=b'exception data')

        # Test GEARMAN_COMMAND_WORK_DATA
        self.command_handler.send_job_data(current_job, b'job data')
        self.assert_sent_command(GEARMAN_COMMAND_WORK_DATA, job_handle=current_job.handle, data=b'job data')

        # Test GEARMAN_COMMAND_WORK_WARNING
        self.command_handler.send_job_warning(current_job, b'job warning')
        self.assert_sent_command(GEARMAN_COMMAND_WORK_WARNING, job_handle=current_job.handle, data=b'job warning')


class WorkerCommandHandlerStateMachineTest(_GearmanAbstractWorkerTest):
    """Test multiple state transitions within a GearmanWorkerCommandHandler

    End to end tests without a server
    """
    connection_manager_class = MockGearmanWorker
    command_handler_class = GearmanWorkerCommandHandler

    def setup_connection_manager(self):
        super(WorkerCommandHandlerStateMachineTest, self).setup_connection_manager()
        self.connection_manager.register_task(b'__test_ability__', None)

    def setup_command_handler(self):
        super(_GearmanAbstractWorkerTest, self).setup_command_handler()
        self.assert_sent_abilities([b'__test_ability__'])
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)

    def test_wakeup_work(self):
        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        self.move_to_state_no_job()

    def test_wakeup_sleep_wakup_work(self):
        self.move_to_state_wakeup()

        self.move_to_state_no_job()

        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        self.move_to_state_no_job()

    def test_multiple_wakeup_then_no_work(self):
        # Awaken the state machine... then give it no work
        self.move_to_state_wakeup()

        for _ in range(5):
            self.command_handler.recv_command(GEARMAN_COMMAND_NOOP)

        self.assert_job_lock(is_locked=True)

        # Pretend like the server has no work... do nothing
        # Moving to state NO_JOB will make sure there's only 1 item on the queue
        self.move_to_state_no_job()

    def test_multiple_work(self):
        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        # After this job completes, we're going to greedily ask for more jobs
        self.move_to_state_no_job()

    def test_worker_already_locked(self):
        other_connection = MockGearmanConnection()
        self.connection_manager.connection_list.append(other_connection)
        self.connection_manager.establish_connection(other_connection)

        other_handler = self.connection_manager.connection_to_handler_map[other_connection]
        other_handler.recv_command(GEARMAN_COMMAND_NOOP)

        # Make sure other handler has a lock
        assert self.connection_manager.command_handler_holding_job_lock == other_handler

        # Make sure OUR handler has nothing incoming
        self.assert_no_pending_commands()

        # Make sure we try to grab a job but fail...so go back to sleep
        self.command_handler.recv_command(GEARMAN_COMMAND_NOOP)
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)

        # Make sure other handler still has lock
        assert self.connection_manager.command_handler_holding_job_lock == other_handler

        # Make the other handler release its lock
        other_handler.recv_command(GEARMAN_COMMAND_NO_JOB)

        # Ensure that the lock has been freed
        self.assert_job_lock(is_locked=False)

        # Try to do work after we have our lock released
        self.move_to_state_wakeup()

        self.move_to_state_job_assign_uniq(self.generate_job_dict())

        self.move_to_state_wakeup()

        self.move_to_state_no_job()

    def move_to_state_wakeup(self):
        self.assert_no_pending_commands()
        self.assert_job_lock(is_locked=False)

        self.command_handler.recv_command(GEARMAN_COMMAND_NOOP)

    def move_to_state_no_job(self):
        """Move us to the NO_JOB state...

        1) We should've most recently sent only a single GEARMAN_COMMAND_GRAB_JOB_UNIQ
        2) We should be awaiting job assignment
        3) Once we receive a NO_JOB, we should say we're going back to sleep"""
        self.assert_awaiting_job()

        self.command_handler.recv_command(GEARMAN_COMMAND_NO_JOB)

        # We should be asleep... which means no pending jobs and we're not awaiting job assignment
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)
        self.assert_no_pending_commands()
        self.assert_job_lock(is_locked=False)

    def move_to_state_job_assign_uniq(self, fake_job):
        """Move us to the JOB_ASSIGN_UNIQ state...

        1) We should've most recently sent only a single GEARMAN_COMMAND_GRAB_JOB_UNIQ
        2) We should be awaiting job assignment
        3) The job we receive should be the one we expected"""
        self.assert_awaiting_job()

        ### NOTE: This recv_command does NOT send out a GEARMAN_COMMAND_JOB_COMPLETE or GEARMAN_COMMAND_JOB_FAIL
        ###           as we're using a MockGearmanConnectionManager with a method that only queues the job
        self.command_handler.recv_command(GEARMAN_COMMAND_JOB_ASSIGN_UNIQ, **fake_job)

        current_job = self.connection_manager.worker_job_queues[self.command_handler].popleft()
        assert current_job.handle == fake_job['job_handle']
        assert current_job.task == fake_job['task']
        assert current_job.unique == fake_job['unique']
        assert current_job.data == fake_job['data']

        # At the end of recv_command(GEARMAN_COMMAND_JOB_ASSIGN_UNIQ)
        self.assert_job_lock(is_locked=False)
        self.assert_sent_command(GEARMAN_COMMAND_PRE_SLEEP)

    def assert_awaiting_job(self):
        self.assert_sent_command(GEARMAN_COMMAND_GRAB_JOB_UNIQ)
        self.assert_no_pending_commands()

    def assert_job_lock(self, is_locked):
        expected_value = (is_locked and self.command_handler) or None
        assert self.connection_manager.command_handler_holding_job_lock == expected_value



#!/usr/bin/env python

import os

from setuptools import setup


def local_file(name):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), name))


SOURCE = local_file('gearman')
README = local_file('README.rst')


# Assignment to placate pyflakes. The actual version is from the exec that
# follows.
__version__ = None

with open(local_file('gearman/version.py')) as o:
    exec(o.read())

assert __version__ is not None


setup(
    name = 'pygearman',
    version = __version__,
    author = 'nerou GmbH',
    author_email = 'info@nerou.de',
    description = 'A Python 3 fork of the Gearman API - Client, worker, and admin client interfaces',
    long_description=open(README).read(),
    url = 'https://github.com/nerou42/pygearman',
    packages = ['gearman'],
    license='Apache / MIT',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- encoding: utf-8

from hypothesis import given
from hypothesis.strategies import sampled_from, sets
import pytest

from gearman.admin_client import GearmanAdminClient, ECHO_STRING
from gearman.admin_client_handler import GearmanAdminClientCommandHandler
from gearman.connection import GearmanConnection
from gearman.errors import GearmanError, InvalidAdminClientState, ProtocolError
from gearman.protocol import (
    GEARMAN_COMMAND_ECHO_RES,
    GEARMAN_COMMAND_ECHO_REQ,
    GEARMAN_SERVER_COMMAND_CANCEL_JOB,
    GEARMAN_SERVER_COMMAND_GETPID,
    GEARMAN_SERVER_COMMAND_SHOW_JOBS,
    GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS,
    GEARMAN_COMMAND_TEXT_COMMAND,
    GEARMAN_SERVER_COMMAND_STATUS, GEARMAN_SERVER_COMMAND_VERSION, GEARMAN_SERVER_COMMAND_WORKERS, GEARMAN_SERVER_COMMAND_MAXQUEUE, GEARMAN_SERVER_COMMAND_SHUTDOWN)

from tests._core_testing import _GearmanAbstractTest, MockGearmanConnectionManager


class MockGearmanAdminClient(GearmanAdminClient, MockGearmanConnectionManager):
    pass


class _StateMachineTest(_GearmanAbstractTest):

    connection_manager_class = MockGearmanAdminClient

    def setUp(self):
        super(_StateMachineTest, self).setUp()
        self.connection_manager.current_connection = self.connection
        self.connection_manager.current_handler = self.command_handler

    def send_server_command(self, expected_command):
        self.command_handler.send_text_command(expected_command)
        expected_line = "%s\n" % expected_command
        self.assert_sent_command(GEARMAN_COMMAND_TEXT_COMMAND, raw_text=expected_line)

        assert self.command_handler._sent_commands[0] == expected_command

    def recv_server_response(self, response_line):
        self.command_handler.recv_command(GEARMAN_COMMAND_TEXT_COMMAND, raw_text=response_line)

    def pop_response(self, expected_command):
        server_cmd, server_response = self.command_handler.pop_response()
        assert expected_command == server_cmd

        return server_response


class CommandHandlerStateMachineTest(_StateMachineTest):
    """Test the public interface a GearmanWorker may need to call in order to update state on a GearmanWorkerCommandHandler"""
    command_handler_class = GearmanAdminClientCommandHandler

    def test_send_illegal_server_commands(self):
        with pytest.raises(ProtocolError):
            self.send_server_command("This is not a server command")

    def test_ping_server(self):
        self.command_handler.send_echo_request(ECHO_STRING)
        self.assert_sent_command(GEARMAN_COMMAND_ECHO_REQ, data=ECHO_STRING)
        assert self.command_handler._sent_commands[0] == GEARMAN_COMMAND_ECHO_REQ

        self.command_handler.recv_command(GEARMAN_COMMAND_ECHO_RES, data=ECHO_STRING)
        server_response = self.pop_response(GEARMAN_COMMAND_ECHO_REQ)
        assert server_response == ECHO_STRING

    def test_state_and_protocol_errors_for_status(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_STATUS)

        # Test premature popping as this we aren't until ready we see the '.'
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_STATUS)

        # Test malformed server status
        with pytest.raises(ProtocolError):
            self.recv_server_response('\t'.join(['12', 'IP-A', 'CLIENT-A']))

        self.recv_server_response('.')

        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_STATUS)
        assert server_response == ()

    def test_response_ready(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_STATUS)

        # We aren't ready until we see the '.'
        assert not self.command_handler.response_ready

        self.recv_server_response(".")
        assert self.command_handler.response_ready

    def test_multiple_status(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_STATUS)
        self.recv_server_response('\t'.join(['test_function', '1', '5', '17']))
        self.recv_server_response('\t'.join(['another_function', '2', '4', '23']))
        self.recv_server_response('.')

        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_STATUS)
        assert len(server_response) == 2

        test_response, another_response = server_response
        assert test_response['task'] == 'test_function'
        assert test_response['queued'] == 1
        assert test_response['running'] == 5
        assert test_response['workers'] == 17

        assert another_response['task'] == 'another_function'
        assert another_response['queued'] == 2
        assert another_response['running'] == 4
        assert another_response['workers'] == 23

    def test_version(self):
        expected_version = '0.12345'

        self.send_server_command(GEARMAN_SERVER_COMMAND_VERSION)
        self.recv_server_response(expected_version)

        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_VERSION)
        assert expected_version == server_response

    def test_state_and_protocol_errors_for_workers(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_WORKERS)

        # Test premature popping as this we aren't until ready we see the '.'
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_WORKERS)

        # Test malformed responses
        with pytest.raises(ProtocolError):
            self.recv_server_response(' '.join(['12', 'IP-A', 'CLIENT-A']))

        with pytest.raises(ProtocolError):
            self.recv_server_response(' '.join(['12', 'IP-A', 'CLIENT-A', 'NOT:']))

        self.recv_server_response('.')

        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_WORKERS)
        assert server_response == ()

    def test_multiple_workers(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_WORKERS)
        self.recv_server_response(' '.join(['12', 'IP-A', 'CLIENT-A', ':', 'function-A', 'function-B']))
        self.recv_server_response(' '.join(['13', 'IP-B', 'CLIENT-B', ':', 'function-C']))
        self.recv_server_response('.')

        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_WORKERS)
        assert len(server_response) == 2

        test_response, another_response = server_response
        assert test_response['file_descriptor'] == '12'
        assert test_response['ip'] == 'IP-A'
        assert test_response['client_id'] == 'CLIENT-A'
        assert test_response['tasks'] == ('function-A', 'function-B')

        assert another_response['file_descriptor'] == '13'
        assert another_response['ip'] == 'IP-B'
        assert another_response['client_id'] == 'CLIENT-B'
        assert another_response['tasks'] == ('function-C', )

    def test_maxqueue(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_MAXQUEUE)
        with pytest.raises(ProtocolError):
            self.recv_server_response('NOT OK')

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_MAXQUEUE)

        self.recv_server_response('OK')
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_MAXQUEUE)
        assert server_response == 'OK'

    def test_shutdown(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHUTDOWN)

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_SHUTDOWN)

        self.recv_server_response(None)
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_SHUTDOWN)
        assert server_response is None

    def test_getpid(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_GETPID)

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_GETPID)

        self.recv_server_response(None)
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_GETPID)
        assert server_response is None

    def test_cancel_job(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_CANCEL_JOB)

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_CANCEL_JOB)

        self.recv_server_response(None)
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_CANCEL_JOB)
        assert server_response is None

    def test_show_jobs_empty(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        self.recv_server_response('.')
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_JOBS)
        assert server_response == ()

    def test_show_jobs_incorrect_tokens(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        with pytest.raises(ProtocolError, match="Received 3 tokens, expected 4 tokens"):
            self.recv_server_response('1\t2\t3')

    def test_show_jobs(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        self.recv_server_response('foo\t1\t2\t3')
        self.recv_server_response('bar\t4\t5\t6')

        # Pop prematurely
        with pytest.raises(InvalidAdminClientState):
            self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        self.recv_server_response(".")
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_JOBS)
        assert server_response == (
            {"handle": "foo", "queued": 1, "canceled": 2, "enabled": 3},
            {"handle": "bar", "queued": 4, "canceled": 5, "enabled": 6},
        )

    def test_show_unique_jobs_empty(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS)

        self.recv_server_response('.')
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS)
        assert server_response == ()

    def test_show_unique_jobs_incorrect_tokens(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS)

        with pytest.raises(ProtocolError, match="Received 3 tokens, expected 1 tokens"):
            self.recv_server_response('1\t2\t3')

    def test_show_unique_jobs(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS)

        self.recv_server_response('foo')
        self.recv_server_response('bar')

        self.recv_server_response(".")
        server_response = self.pop_response(GEARMAN_SERVER_COMMAND_SHOW_UNIQUE_JOBS)
        assert server_response == (
            {"unique": ["foo"]},
            {"unique": ["bar"]},
        )

    def test_recv_server_response_without_command(self):
        with pytest.raises(InvalidAdminClientState, match='Received an unexpected server response'):
            self.recv_server_response(b'123')


def _assert_gearman_connection_equal(conn1, conn2):
    # There's currently no equality method defined on ``GearmanConnection``,
    # so separate instances are unequal.
    #
    # Adding an equality method has a bunch of other subtle side effects
    # (e.g. related to hashing), so making this method test-only is safer
    # until we have better test coverage.
    #
    assert isinstance(conn1, GearmanConnection)
    assert isinstance(conn2, GearmanConnection)
    assert conn1.gearman_host == conn2.gearman_host
    assert conn1.gearman_port == conn2.gearman_port
    assert conn1.keyfile == conn2.keyfile
    assert conn1.certfile == conn2.certfile
    assert conn1.ca_certs == conn2.ca_certs


class TestGearmanAdminClient(object):

    @pytest.mark.parametrize('host_config', [
        # As a string specifying a port
        "localhost:4730",

        # If a port isn't specified, we use the default port
        "localhost",

        # As a tuple (with and without int argument)
        ("localhost", "4730"),
        ("localhost", 4730),
    ])
    def test_creating_with_host_port_pair_str(self, host_config):
        client = GearmanAdminClient(host_list=[host_config])
        assert len(client.connection_list) == 1

        _assert_gearman_connection_equal(
            conn1=GearmanConnection(host="localhost", port=4730),
            conn2=client.connection_list[0]
        )

    def test_multiple_hosts_is_error(self):
        with pytest.raises(GearmanError, match="Only pass a single host"):
            GearmanAdminClient(host_list=[
                "localhost:1234",
                "example.org:5678",
            ])

    @given(sets(
        sampled_from(['host', 'port', 'keyfile', 'certfile', 'ca_certs']),
        min_size=1))
    def test_passing_incomplete_ssl_data_is_error(self, keys_to_delete):
        host_config = {
            "host": "localhost",
            "port": 5000,
            "keyfile": "key.txt",
            "certfile": "cert.txt",
            "ca_certs": "ca_certs.txt",
        }

        for key in keys_to_delete:
            del host_config[key]

        with pytest.raises(GearmanError, match="Incomplete SSL connection definition"):
            GearmanAdminClient(host_list=[host_config])

    @pytest.mark.parametrize('bad_config', [None, 1, object])
    def test_passing_bad_host_config_is_error(self, bad_config):
        with pytest.raises(GearmanError, match="Expected str, tuple or dict"):
            GearmanAdminClient(host_list=[bad_config])

    def test_creating_with_ssl_host(self):
        host_config = {
            "host": "localhost",
            "port": 5000,
            "keyfile": "key.txt",
            "certfile": "cert.txt",
            "ca_certs": "ca_certs.txt",
        }

        client = GearmanAdminClient(host_list=[host_config])
        assert len(client.connection_list) == 1

        _assert_gearman_connection_equal(
            conn1=GearmanConnection(host="localhost", port=5000, keyfile="key.txt", certfile="cert.txt", ca_certs="ca_certs.txt"),
            conn2=client.connection_list[0]
        )


class TestGearmanAdminClientCommandHandler:

    command_handler = GearmanAdminClientCommandHandler(
        connection_manager=MockGearmanAdminClient()
    )

    def test_decodes_data_correctly(self):
        assert self.command_handler.decode_data(b"123") == b"123"

    def test_encodes_data_correctly(self):
        assert self.command_handler.encode_data(b"123") == b"123"


class BrokenCommandHandler(GearmanAdminClientCommandHandler):
    recv_server_show_jobs = None


class TestBrokenCommandHandlerStateMachine(_StateMachineTest):
    """
    This exists to hit a line that isn't reachable in normal operation.
    """
    command_handler_class = BrokenCommandHandler

    def test_show_jobs_empty(self):
        self.send_server_command(GEARMAN_SERVER_COMMAND_SHOW_JOBS)

        with pytest.raises(ValueError, match="Could not handle command: 'show_jobs'"):
            self.recv_server_response('.')



"""
Gearman API - Client, worker, and admin client interfaces
"""

from gearman.admin_client import GearmanAdminClient
from gearman.client import GearmanClient
from gearman.version import __version__  # noqa
from gearman.worker import GearmanWorker

from gearman.connection_manager import DataEncoder
from gearman.constants import PRIORITY_NONE, PRIORITY_LOW, PRIORITY_HIGH, JOB_PENDING, JOB_CREATED, JOB_FAILED, JOB_COMPLETE, JOB_UNKNOWN

__all__ = [
    "GearmanAdminClient",
    "GearmanClient",
    "GearmanWorker",

    "DataEncoder",

    "PRIORITY_NONE",
    "PRIORITY_LOW",
    "PRIORITY_HIGH",
    "JOB_PENDING",
    "JOB_CREATED",
    "JOB_FAILED",
    "JOB_COMPLETE",
    "JOB_UNKNOWN",

    "__version__",
]



# -*- encoding: utf-8

import array
import struct

from hypothesis import given
from hypothesis.strategies import binary
import pytest

from gearman import compat, protocol
from gearman.errors import ProtocolError
from tests._core_testing import _GearmanAbstractTest


class FakeType(object):
    pass


class TestProtocolBinaryCommands(object):
    #######################
    # Begin parsing tests #
    #######################
    @pytest.mark.parametrize('data,is_response', [
        (b"DDDDAAAABBBBCCCC", True),
        (b"%sAAAABBBBCCCC" % protocol.MAGIC_RES_STRING, False),
        (b"%sAAAABBBBCCCC" % protocol.MAGIC_REQ_STRING, True),
    ])
    def test_parsing_errors_raw(self, data, is_response):
        in_buffer = array.array("b", data)
        with pytest.raises(ProtocolError):
            protocol.parse_binary_command(in_buffer, is_response=is_response)

    @pytest.mark.parametrize('struct_components', [
        # Raise unknown command errors
        (("!4sII", protocol.MAGIC_RES_STRING, 1234, 0)),

        # Raise an error on our imaginary GEARMAN_COMMAND_TEXT_COMMAND
        (("!4sII4s", protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_TEXT_COMMAND, 4, b'ABCD')),

        # Raise an error on receiving an unexpected payload
        (("!4sII4s", protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_NOOP, 4, b'ABCD')),

        # Raise an error when receiving some invalid Gearman data
        (("!4sII", protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_SUBMIT_JOB, 0)),
    ])
    def test_parsing_errors_struct(self, struct_components):
        data = struct.pack(*struct_components)
        in_buffer = array.array("b", data)
        with pytest.raises(ProtocolError):
            protocol.parse_binary_command(in_buffer)

    @given(binary())
    def test_parsing_either_succeeds_or_is_protocolerror(self, in_buffer):
        try:
            protocol.parse_binary_command(array.array("b", in_buffer))
        except ProtocolError:
            pass

    def test_parsing_request(self):
        # Test parsing a request for a job (server side parsing)
        grab_job_command_buffer = struct.pack('!4sII', protocol.MAGIC_REQ_STRING, protocol.GEARMAN_COMMAND_GRAB_JOB_UNIQ, 0)
        grab_job_command_buffer = array.array("b", grab_job_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(grab_job_command_buffer, is_response=False)
        assert cmd_type == protocol.GEARMAN_COMMAND_GRAB_JOB_UNIQ
        assert cmd_args == {}
        assert cmd_len == len(grab_job_command_buffer)

    def test_parsing_without_enough_data(self):
        # Test that we return with nothing to do... received a partial packet
        not_enough_data_command_buffer = struct.pack('!4s', protocol.MAGIC_RES_STRING)
        not_enough_data_command_buffer = array.array("b", not_enough_data_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(not_enough_data_command_buffer)
        assert cmd_type is None
        assert cmd_args is None
        assert cmd_len == 0

        # Test that we return with nothing to do... received a partial packet (expected binary payload of size 4, got 0)
        not_enough_data_command_buffer = struct.pack('!4sII', protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_ECHO_RES, 4)
        not_enough_data_command_buffer = array.array("b", not_enough_data_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(not_enough_data_command_buffer)
        assert cmd_type is None
        assert cmd_args is None
        assert cmd_len == 0

    def test_parsing_no_args(self):
        noop_command_buffer = struct.pack('!4sII', protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_NOOP, 0)
        noop_command_buffer = array.array("b", noop_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(noop_command_buffer)
        assert cmd_type == protocol.GEARMAN_COMMAND_NOOP
        assert cmd_args == dict()
        assert cmd_len == len(noop_command_buffer)

    def test_parsing_single_arg(self):
        echoed_string = b'abcd'
        echo_command_buffer = struct.pack('!4sII4s', protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_ECHO_RES, 4, echoed_string)
        echo_command_buffer = array.array("b", echo_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(echo_command_buffer)
        assert cmd_type == protocol.GEARMAN_COMMAND_ECHO_RES
        assert cmd_args == {u"data": echoed_string}
        assert cmd_len == len(echo_command_buffer)

    def test_parsing_single_arg_with_extra_data(self):
        echoed_string = b'abcd'
        excess_bytes = 5
        excess_data = echoed_string + (protocol.NULL_CHAR * excess_bytes)
        excess_echo_command_buffer = struct.pack('!4sII9s', protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_ECHO_RES, 4, excess_data)
        excess_echo_command_buffer = array.array("b", excess_echo_command_buffer)

        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(excess_echo_command_buffer)
        assert cmd_type == protocol.GEARMAN_COMMAND_ECHO_RES
        assert cmd_args == dict(data=echoed_string)
        assert cmd_len == len(excess_echo_command_buffer) - excess_bytes

    def test_parsing_multiple_args(self):
        # Tests ordered argument processing and proper NULL_CHAR splitting
        expected_data = protocol.NULL_CHAR * 4
        binary_payload = protocol.NULL_CHAR.join([b'test', b'function', b'identifier', expected_data])
        payload_size = len(binary_payload)

        uniq_command_buffer = struct.pack('!4sII%ds' % payload_size, protocol.MAGIC_RES_STRING, protocol.GEARMAN_COMMAND_JOB_ASSIGN_UNIQ, payload_size, binary_payload)
        uniq_command_buffer = array.array("b", uniq_command_buffer)
        cmd_type, cmd_args, cmd_len = protocol.parse_binary_command(uniq_command_buffer)
        assert cmd_type == protocol.GEARMAN_COMMAND_JOB_ASSIGN_UNIQ
        assert cmd_args == {
            u"job_handle": b"test",
            u"task": b"function",
            u"unique": b"identifier",
            u"data": expected_data
        }
        assert cmd_len == len(uniq_command_buffer)

    #######################
    # Begin packing tests #
    #######################
    @pytest.mark.parametrize('cmd_type,cmd_args', [
        # Assert we get an unknown command
        (1234, {}),

        # Assert we get a fake command
        (protocol.GEARMAN_COMMAND_TEXT_COMMAND, {}),

        # Assert we get arg mismatch, got 1, expecting 0
        (protocol.GEARMAN_COMMAND_GRAB_JOB, {u"extra": u"arguments"}),

        # Assert we get arg mismatch, got 0, expecting 1
        (protocol.GEARMAN_COMMAND_JOB_CREATED, {}),

        # Assert we get arg mismatch (name), got 1, expecting 1
        (protocol.GEARMAN_COMMAND_JOB_CREATED, {u"extra": "arguments"}),

        # Assert we get a non-string argument
        (protocol.GEARMAN_COMMAND_JOB_CREATED, {u"job_handle": 12345}),

        # Assert we get a non-binary argument with any other non-binary type.
        (protocol.GEARMAN_COMMAND_JOB_CREATED, {u"job_handle": FakeType()}),

        # Assert we check for NULLs in all but the "last" argument,
        # where last depends on the cmd_type.
        (protocol.GEARMAN_COMMAND_SUBMIT_JOB,
         {u"task": b"funct\x00ion", u"data": b"abcd", u"unique": b"12345"}),
        (protocol.GEARMAN_COMMAND_SUBMIT_JOB,
         {u"task": b"function", u"data": b"abcd", u"unique": b"123\x00\x0045"}),
    ])
    def test_packing_errors(self, cmd_type, cmd_args):
        with pytest.raises(ProtocolError):
            protocol.pack_binary_command(cmd_type=cmd_type, cmd_args=cmd_args)

    def test_packing_errors_last_arg(self):
        # Assert we check for NULLs in all but the "last" argument,
        # where last depends on the cmd_type.
        protocol.pack_binary_command(
            cmd_type=protocol.GEARMAN_COMMAND_SUBMIT_JOB,
            cmd_args={u"task": b"function", u"data": b"ab\x00cd", u"unique": b"12345"}
        )

    def test_packing_response(self):
        # Test packing a response for a job (server side packing)
        cmd_type = protocol.GEARMAN_COMMAND_NO_JOB
        cmd_args = dict()

        expected_command_buffer = struct.pack('!4sII', protocol.MAGIC_RES_STRING, cmd_type, 0)
        packed_command_buffer = protocol.pack_binary_command(cmd_type, cmd_args, is_response=True)
        assert packed_command_buffer == expected_command_buffer

    def test_packing_no_arg(self):
        cmd_type = protocol.GEARMAN_COMMAND_NOOP
        cmd_args = dict()

        expected_command_buffer = struct.pack('!4sII', protocol.MAGIC_REQ_STRING, cmd_type, 0)
        packed_command_buffer = protocol.pack_binary_command(cmd_type, cmd_args)
        assert packed_command_buffer == expected_command_buffer

    @pytest.mark.parametrize('cmd_args, cmd_resp', [
        ({u"data": b"abcde"}, {u"data": b"abcde"}),
        ({u"data": "abcde"}, {u"data": b"abcde"}),
        ({u"data": u"abcde"}, {u"data": b"abcde"}),
    ])
    def test_packing_single_arg(self, cmd_args, cmd_resp):
        cmd_type = protocol.GEARMAN_COMMAND_ECHO_REQ

        expected_payload_size = len(cmd_args['data'])
        expected_format = '!4sII%ds' % expected_payload_size

        expected_command_buffer = struct.pack(
            expected_format,
            protocol.MAGIC_REQ_STRING,
            cmd_type,
            expected_payload_size,
            cmd_resp['data'],
        )

        packed_command_buffer = protocol.pack_binary_command(cmd_type, cmd_args)
        assert packed_command_buffer == expected_command_buffer

    def test_packing_multiple_args(self):
        cmd_type = protocol.GEARMAN_COMMAND_SUBMIT_JOB
        cmd_args = dict(task=b'function', unique=b'12345', data=b'abcd')

        ordered_parameters = [cmd_args['task'], cmd_args['unique'], cmd_args['data']]

        expected_payload = protocol.NULL_CHAR.join(ordered_parameters)
        expected_payload_size = len(expected_payload)
        expected_format = '!4sII%ds' % expected_payload_size
        expected_command_buffer = struct.pack(expected_format, protocol.MAGIC_REQ_STRING, cmd_type, expected_payload_size, expected_payload)

        packed_command_buffer = protocol.pack_binary_command(cmd_type, cmd_args)
        assert packed_command_buffer == expected_command_buffer


class TestProtocolTextCommands(object):
    #######################
    # Begin parsing tests #
    #######################
    def test_parsing_errors(self):
        received_data = array.array("b", b"Hello\x00there\n")
        with pytest.raises(ProtocolError):
            protocol.parse_text_command(received_data)

    def test_parsing_without_enough_data(self):
        received_data = array.array("b", b"Hello there")
        cmd_type, cmd_response, cmd_len = protocol.parse_text_command(received_data)
        assert cmd_type is None
        assert cmd_response is None
        assert cmd_len == 0

    def test_parsing_single_line(self):
        received_data = array.array("b", b"Hello there\n")
        cmd_type, cmd_response, cmd_len = protocol.parse_text_command(received_data)
        assert cmd_type == protocol.GEARMAN_COMMAND_TEXT_COMMAND
        assert cmd_response == {
            u"raw_text": compat.array_to_bytes(received_data).strip()
        }
        assert cmd_len == len(received_data)

    def test_parsing_multi_line(self):
        sentence_one = array.array("b", b"Hello there\n")
        sentence_two = array.array("b", b"My name is bob\n")
        received_data = sentence_one + sentence_two

        cmd_type, cmd_response, cmd_len = protocol.parse_text_command(received_data)
        assert cmd_type == protocol.GEARMAN_COMMAND_TEXT_COMMAND
        assert cmd_response == {
            u"raw_text": compat.array_to_bytes(sentence_one).strip()
        }
        assert cmd_len == len(sentence_one)

    @pytest.mark.parametrize('cmd_type,cmd_args', [
        # Test bad command type
        (protocol.GEARMAN_COMMAND_NOOP, {}),

        # Test missing args
        (protocol.GEARMAN_COMMAND_TEXT_COMMAND, {}),

        # Test misnamed parameter dict
        (protocol.GEARMAN_COMMAND_TEXT_COMMAND, {u"bad_text": u"abcdefghij"}),
    ])
    def test_packing_errors(self, cmd_type, cmd_args):
        with pytest.raises(ProtocolError):
            protocol.pack_text_command(cmd_type=cmd_type, cmd_args=cmd_args)

    #######################
    # Begin packing tests #
    #######################
    def test_packing_single_line(self):
        expected_string = 'Hello world'
        cmd_type = protocol.GEARMAN_COMMAND_TEXT_COMMAND
        cmd_args = {u"raw_text": expected_string}

        packed_command = protocol.pack_text_command(cmd_type, cmd_args)
        assert packed_command == expected_string


class GearmanCommandHandlerTest(_GearmanAbstractTest):
    """Tests the base CommandHandler class that underpins all other CommandHandlerTests"""
    def _test_recv_command(self):
        # recv_echo_res and recv_error are predefined on the CommandHandler
        self.command_handler.recv_command(protocol.GEARMAN_COMMAND_NOOP)
        self.assert_recv_command(protocol.GEARMAN_COMMAND_NOOP)

        # The mock handler never implemented 'recv_all_yours' so we should get an attribute error here
        with pytest.raises(ValueError):
            self.command_handler.recv_command(protocol.GEARMAN_COMMAND_ALL_YOURS)

    def _test_send_command(self):
        self.command_handler.send_command(protocol.GEARMAN_COMMAND_NOOP)
        self.assert_sent_command(protocol.GEARMAN_COMMAND_NOOP)

        # The mock handler never implemented 'recv_all_yours' so we should get an attribute error here
        self.command_handler.send_command(protocol.GEARMAN_COMMAND_ECHO_REQ, text='hello world')
        self.assert_sent_command(protocol.GEARMAN_COMMAND_ECHO_REQ, text='hello world')

    def assert_recv_command(self, expected_cmd_type, **expected_cmd_args):
        cmd_type, cmd_args = self.command_handler.recv_command_queue.popleft()
        self.assert_commands_equal(cmd_type, expected_cmd_type)
        assert cmd_args == expected_cmd_args

    def assert_sent_command(self, expected_cmd_type, **expected_cmd_args):
        # All commands should be sent via the CommandHandler
        handler_cmd_type, handler_cmd_args = self.command_handler.sent_command_queue.popleft()
        self.assert_commands_equal(handler_cmd_type, expected_cmd_type)
        assert handler_cmd_args == expected_cmd_args

        super(GearmanCommandHandlerTest, self).assert_sent_command(expected_cmd_type, **expected_cmd_args)
